from fastapi import FastAPI, Depends, Path
from fastapi_namespace.namespace import Namespace, Resource
from fastapi_namespace.mixins.token import OpaqueTokenMixin

async def test(b: str, asd: str = Path(...), c:str = "") -> bool:
    print(asd, b, c)
    return 1

namespace = Namespace(
    prefix="/abcv",
    dependencies=[Depends(test)]
)

from fastapi import FastAPI
from pydantic import BaseModel
from fastapi_namespace.mixins import MixinBase

def depend_test4(last: str) -> bool:
    print(last)

class Test(MixinBase):

    global_dependencies = [Depends(depend_test4)]

def depend_test5(last2: str) -> bool:
    print(last2)
class Test2(Test):
    check = 4567
    def __init__(self):
        self.add_global_depends(Depends(self.getAge))

    def getAge(self, age: int):
        print('454545', age)
        print(self.check)

class ItemResponse(BaseModel):
    id: str




app = FastAPI()




def depend_test(arg: str, arg2: str) -> bool:
    print(arg, arg2)

def depend_test2(arg3: str, arg4: str) -> bool:
    print(arg3, arg4)

def depend_get(get: str) -> bool:
    print(get)

def depend_post(post: str) -> bool:
    print(post)

from fastapi import Security

@namespace.route('/{asd}')
class Root(OpaqueTokenMixin):

    global_dependencies = [Depends(depend_test), Depends(depend_test2)]
    get_dependencies = [Depends(depend_get)]
    post_dependencies = [Depends(depend_post)]
    def get(self) -> ItemResponse:
        return ItemResponse(id="1234")

    async def post(self) -> ItemResponse:
        return ItemResponse(id="1234")


app.include_router(namespace)


if __name__ == "__main__":
    from uvicorn import run
    run(
        app=app,
        port=23457
    )



from typing import Literal, Callable, Any, NewType, Protocol, Sequence
from fastapi.params import Depends

MethodType_ = Literal["get", "post", "put", "delete", 'options', 'head', 'patch', 'trace']
MethodType: MethodType_ = NewType('MethodType', MethodType_)
MethodHandler = Callable[[Any], Any]
DependsType = Literal[
    'global_dependencies',
    'get_dependencies',
    'post_dependencies',
    'put_dependencies',
    'delete_dependencies',
    'options_dependencies',
    'head_dependencies',
    'patch_dependencies',
    'trace_dependencies'
]
class ResourceProtocol(Protocol):
    global_dependencies: Sequence[Depends]
    get_dependencies: Sequence[Depends]
    post_dependencies: Sequence[Depends]
    put_dependencies: Sequence[Depends]
    delete_dependencies: Sequence[Depends]
    options_dependencies: Sequence[Depends]
    head_dependencies: Sequence[Depends]
    patch_dependencies: Sequence[Depends]
    trace_dependencies: Sequence[Depends]


import string
import secrets

def delete_none(dict_) -> dict:
    return {k: v for k, v in dict_.items() if v is not None}


alphabet = string.ascii_letters
def gen_op_id(length: int = 15) -> str:
    return ''.join(secrets.choice(alphabet) for i in range(length))


from abc import ABCMeta, abstractmethod
from typing import Iterable, Callable, Any, Literal
from fastapi.params import Depends
from inspect import (
    Parameter,
    Signature,
    signature,
    iscoroutinefunction
)
from .typings import MethodHandler, MethodType, ResourceProtocol
from .utils import gen_op_id

def resource_dependant_name(key: str) -> str:
    return f'{key}_dependencies'


class Resource:
    global_dependencies: Iterable[Depends]
    get_dependencies: Iterable[Depends]
    post_dependencies: Iterable[Depends]
    put_dependencies: Iterable[Depends]
    delete_dependencies: Iterable[Depends]
    options_dependencies: Iterable[Depends]
    head_dependencies: Iterable[Depends]
    patch_dependencies: Iterable[Depends]
    trace_dependencies: Iterable[Depends]

    def get_dependant(self, method_handler, depends: Depends) -> Callable:
        assert isinstance(depends, Depends), "Dependant must be of type Depends!"
        method_handler_signature = signature(method_handler)
        method_handler_parameters = method_handler_signature.parameters

        op_id = gen_op_id()

        def wrap(**kwargs):
            kwargs.pop(op_id, None)
            return method_handler(**kwargs)

        async def async_wrap(**kwargs):
            kwargs.pop(op_id, None)
            return await method_handler(**kwargs)

        params = [
            Parameter(name=op_id, kind=Parameter.POSITIONAL_OR_KEYWORD, default=depends),
            *method_handler_parameters.values()
        ]

        if iscoroutinefunction(method_handler):
            async_wrap.__signature__ = Signature(
                parameters=params,
                return_annotation=method_handler_signature.return_annotation
            )
            return async_wrap
        else:
            wrap.__signature__ = Signature(
                parameters=params,
                return_annotation=method_handler_signature.return_annotation
            )
            return wrap

    def get_method_handler(
            self,
            method_handler,
    ) -> MethodHandler:
        method_func = method_handler
        method_name: MethodType = method_handler.__name__

        method_dependencies = getattr(self, resource_dependant_name(method_name), [])
        method_dependencies = [*method_dependencies]
        method_dependencies.reverse()
        global_dependencies = getattr(self, resource_dependant_name('global'), [])
        global_dependencies = [*global_dependencies]
        global_dependencies.reverse()

        dependencies = [*method_dependencies, *global_dependencies]

        for depends in dependencies:
            method_func = self.get_dependant(method_func, depends)
        return method_func


from starlette.routing import BaseRoute
from fastapi import Depends, Response
from fastapi.types import IncEx
from fastapi.routing import APIRoute

from typing import TypedDict, NotRequired, Sequence, Type, Any, Callable, TypeVar
from enum import Enum

class MethodDocument(TypedDict):
    response_model: NotRequired[Any]
    status_code: NotRequired[int]
    tags: NotRequired[list[str | Enum]]
    dependencies: NotRequired[Sequence[Depends]]
    summary: NotRequired[str]
    description: NotRequired[str]
    response_description: NotRequired[str]
    responses: NotRequired[dict[[int | str], dict[str, Any]]]
    deprecated: NotRequired[bool]
    operation_id: NotRequired[str]
    response_model_include: NotRequired[IncEx]
    response_model_exclude: NotRequired[IncEx]
    response_model_by_alias: NotRequired[bool]
    response_model_exclude_unset: NotRequired[bool]
    response_model_exclude_defaults: NotRequired[bool]
    response_model_exclude_none: NotRequired[bool]
    include_in_schema: NotRequired[bool]
    response_class: NotRequired[Type[Response]]
    name: NotRequired[str]
    callbacks: NotRequired[list[BaseRoute]]
    openapi_extra: NotRequired[dict[str, Any]]
    generate_unique_id_function: NotRequired[Callable[[APIRoute], str]]


DecoratedCallable = TypeVar("DecoratedCallable", bound=Callable[..., Any])


from starlette.routing import BaseRoute, Match
from starlette.types import ASGIApp, Lifespan, Receive, Scope, Send
from starlette.datastructures import URL
from starlette.responses import RedirectResponse
from starlette._utils import get_route_path
from fastapi import APIRouter
from fastapi.types import (
    IncEx,
)
from fastapi.responses import JSONResponse as ORJSONResponse
from fastapi.routing import APIRoute
from fastapi.utils import generate_unique_id
from fastapi import Response
from fastapi.params import Depends

from .resource import Resource
from .types import (
    MethodDocument,
    DecoratedCallable
)
from typing import (
    Optional,
    Callable,
    Sequence,
    Type,
    Any,
    Literal,
)
from .utils import delete_none
from enum import Enum
"""
    실행순서
    route out
    doc out
    doc in
    route in 
    resource set doc

"""

__methods__ = ['get', 'post', 'put', 'delete', 'options', 'head', 'patch', 'trace']


class Namespace(APIRouter):
    def __init__(
            self,
            *,
            prefix: str | None = "",
            tags: list[str | Enum] | None = None,
            dependencies: Sequence[Depends] = None,
            default_response_class: Type[Response] = ORJSONResponse,
            responses: dict[int | str, dict[str, Any]] | None = None,
            callbacks: list[BaseRoute] | None = None,
            routes: list[BaseRoute] | None = None,
            redirect_slashes: bool = True,
            default: ASGIApp | None = None,
            dependency_overrides_provider: Any | None = None,
            route_class: Type[APIRoute] = APIRoute,
            lifespan: Lifespan[Any] | None = None,
            deprecated: bool | None = None,
            include_in_schema: bool = True,
            generate_unique_id_function: Callable[[APIRoute], str] = generate_unique_id,
    ):
        super().__init__(
            prefix=prefix,
            tags=tags,
            dependencies=dependencies,
            default_response_class=default_response_class,
            responses=responses,
            callbacks=callbacks,
            routes=routes,
            redirect_slashes=redirect_slashes,
            default=default,
            dependency_overrides_provider=dependency_overrides_provider,
            route_class=route_class,
            lifespan=lifespan,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
            generate_unique_id_function=generate_unique_id_function
        )
        self.get = None
        self.post = None
        self.put = None
        self.delete = None
        self.options = None
        self.head = None
        self.patch = None
        self.trace = None

    def route(
            self,
            path,
            *,
            response_model: Any = None,
            status_code: int | None = None,
            tags: Optional[list[str | Enum]] = None,
            dependencies: Sequence[Depends] | None = None,
            summary: str | None = None,
            description: str | None = None,
            response_description: str = "Successful Response",
            responses: dict[[int | str], dict[str, Any]] | None = None,
            deprecated: bool | None = None,
            operation_id: str | None = None,
            response_model_include: IncEx | None = None,
            response_model_exclude: IncEx | None = None,
            response_model_by_alias: bool = True,
            response_model_exclude_unset: bool = False,
            response_model_exclude_defaults: bool = False,
            response_model_exclude_none: bool = False,
            include_in_schema: bool = True,
            response_class: Type[Response] = ORJSONResponse,
            name: str | None = None,
            callbacks: list[BaseRoute] | None = None,
            openapi_extra: dict[str, Any] | None = None,
            generate_unique_id_function: Callable[[APIRoute], str] = generate_unique_id,
    ) -> Callable[[Resource], Resource]:
        """
        FastAPI APIRoute Class에 deprecated 된 route 함수를 overwriting 하여 사용

        Args:
            path:
            response_model:
            status_code:
            tags: namespace.doc 사용시 append 됨
            dependencies: namespace.doc 사용시 append 됨
            summary:
            description:
            response_description:
            responses:
            deprecated:
            operation_id:
            response_model_include:
            response_model_exclude:
            response_model_by_alias:
            response_model_exclude_unset:
            response_model_exclude_defaults:
            response_model_exclude_none:
            include_in_schema:
            response_class:
            name:
            callbacks: namespace.doc 사용시 append 됨
            openapi_extra:
            generate_unique_id_function:
        """
        def wrap(class_: Resource) -> Resource:
            instance = class_()
            assert isinstance(instance, Resource), "Instance must be of type Resource"
            for meth in __methods__:
                if not hasattr(instance, meth):
                    continue

                meth_func = getattr(instance, meth)
                self._add_method(
                    path,
                    meth,
                    meth_func,
                    response_model=response_model,
                    status_code=status_code,
                    tags=tags,
                    dependencies=dependencies,
                    summary=summary,
                    description=description,
                    response_description=response_description,
                    responses=responses,
                    deprecated=deprecated,
                    operation_id=operation_id,
                    response_model_include=response_model_include,
                    response_model_exclude=response_model_exclude,
                    response_model_by_alias=response_model_by_alias,
                    response_model_exclude_unset=response_model_exclude_unset,
                    response_model_exclude_defaults=response_model_exclude_defaults,
                    response_model_exclude_none=response_model_exclude_none,
                    include_in_schema=include_in_schema,
                    response_class=response_class,
                    name=name,
                    callbacks=callbacks,
                    openapi_extra=openapi_extra,
                    generate_unique_id_function=generate_unique_id_function
                )
            return class_
        return wrap

    def _add_method(
            self,
            path,
            method: Literal['get', 'post', 'put', 'delete', 'options', 'head', 'patch', 'trace'],
            func,
            *,
            response_model: Any = None,
            status_code: int | None = None,
            tags: Optional[list[str | Enum]] = None,
            dependencies: Sequence[Depends] | None = None,
            summary: str | None = None,
            description: str | None = None,
            response_description: str = "Successful Response",
            responses: dict[[int | str], dict[str, Any]] | None = None,
            deprecated: bool | None = None,
            operation_id: str | None = None,
            response_model_include: IncEx | None = None,
            response_model_exclude: IncEx | None = None,
            response_model_by_alias: bool = True,
            response_model_exclude_unset: bool = False,
            response_model_exclude_defaults: bool = False,
            response_model_exclude_none: bool = False,
            include_in_schema: bool = True,
            response_class: Type[Response] = ORJSONResponse,
            name: str | None = None,
            callbacks: list[BaseRoute] | None = None,
            openapi_extra: dict[str, Any] | None = None,
            generate_unique_id_function: Callable[[APIRoute], str] = generate_unique_id,
    ):
        kwargs: MethodDocument = MethodDocument(
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            dependencies=dependencies,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            operation_id=operation_id,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            include_in_schema=include_in_schema,
            response_class=response_class,
            name=name,
            callbacks=callbacks,
            openapi_extra=openapi_extra,
            generate_unique_id_function=generate_unique_id_function
        )
        #
        if hasattr(func.__func__, "__meth_doc__"):
            doc: MethodDocument = func.__func__.__meth_doc__
            if doc.get("dependencies") is not None and dependencies is not None:
                doc["dependencies"] = [*dependencies, *doc["dependencies"]]
            if doc.get("callbacks") is not None and callbacks is not None:
                doc["callbacks"] = [*callbacks, *doc["callbacks"]]
            if doc.get("tags") is not None and tags is not None:
                doc["tags"] = [*tags, *doc["tags"]]
            kwargs.update({
                **doc,
            })
        default_summary = f"{func.__self__.__class__.__name__}_{func.__name__}"
        kwargs.update({
            "summary": default_summary if (summary := kwargs.get("summary", None)) is None else f"{summary} {default_summary}"
        })
        func.__func__.__meth_doc__ = delete_none(kwargs)

        new_func = func.__self__.get_method_handler(
            func
        )
        self.add_api_route(
            path=path,
            endpoint=new_func,
            methods=[method.upper()],
            **func.__func__.__meth_doc__
        )

    def doc(
            self,
            summary: str | None = None,
            description: str | None = None,
            name: str | None = None,
            tags: Optional[list[str | Enum]] = None,
            dependencies: Sequence[Depends] | None = None,
            callbacks: list[BaseRoute] | None = None,
            response_model: Any = None,
            status_code: int | None = None,
            response_description: str = "Successful Response",
            responses: dict[[int | str], dict[str, Any]] | None = None,
            deprecated: bool | None = None,
            operation_id: str | None = None,
            response_model_include: IncEx | None = None,
            response_model_exclude: IncEx | None = None,
            response_model_by_alias: bool = True,
            response_model_exclude_unset: bool = False,
            response_model_exclude_defaults: bool = False,
            response_model_exclude_none: bool = False,
            include_in_schema: bool = True,
            response_class: Type[Response] = ORJSONResponse,
            openapi_extra: dict[str, Any] | None = None,
            generate_unique_id_function: Callable[[APIRoute], str] = generate_unique_id,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        tags = tags or []
        dependencies = dependencies or []
        callbacks = callbacks or []

        def wrap(func: DecoratedCallable) -> DecoratedCallable:
            func.__meth_doc__ = MethodDocument(
                response_model=response_model,
                status_code=status_code,
                tags=tags,
                dependencies=dependencies,
                summary=summary,
                description=description,
                response_description=response_description,
                responses=responses,
                deprecated=deprecated,
                operation_id=operation_id,
                response_model_include=response_model_include,
                response_model_exclude=response_model_exclude,
                response_model_by_alias=response_model_by_alias,
                response_model_exclude_unset=response_model_exclude_unset,
                response_model_exclude_defaults=response_model_exclude_defaults,
                response_model_exclude_none=response_model_exclude_none,
                include_in_schema=include_in_schema,
                response_class=response_class,
                name=name,
                callbacks=callbacks,
                openapi_extra=openapi_extra,
                generate_unique_id_function=generate_unique_id_function,
            )
            return func
        return wrap

    @staticmethod
    def get_method_func_name(func) -> str:
        return f"{func.__class__.__name__}_{func.__name__}"



from .tokenBaseMixin import TokenBaseMixin, TokenValidator, TokenInfoValidator, TI
from .typings import OpaqueTokenInfo, OpaqueToken, TokenInfo, RawToken, Token
from .utils import validate_opaque_token, validate_opaque_token_info

from secrets import token_urlsafe
from uuid import uuid4

class OpaqueTokenMixin(TokenBaseMixin[OpaqueToken, OpaqueTokenInfo]):

    def __init__(self):
        super().__init__(
            token=OpaqueToken,
            token_info=OpaqueTokenInfo,
            token_validator=validate_opaque_token,
            token_info_validator=validate_opaque_token_info
        )

    def create_token(self) -> RawToken:
        return token_urlsafe(48)

    def _make_token(self, token: Token) -> OpaqueTokenInfo:
        return OpaqueTokenInfo(
            **token,
            idf=uuid4().hex
        )



from redis.asyncio import Redis
from redis.asyncio.client import Pipeline

AsyncRedis = Redis
AsyncPipeline = Pipeline

from typing import (
    Literal,
    NewType,
    Any,
    Union,
    Generic
)
from typing_extensions import (
    TypedDict,
    NotRequired
)


TokenKey = NewType('TokenKey', str)
UserTokenKey = NewType('UserTokenKey', str)

MethodType = Literal["get", "post", "put", "delete", "patch", "head", "options", "trace"]
TokenType = Literal["ACCESS", "REFRESH"]
# RawToken = NewType("RawToken", str)
RawToken = str
# TokenLimit = NewType("TokenLimit", int)
TokenLimit = int
# TokenExpire = NewType("TokenExpire", int)
TokenExpire = int
# TokenIdentify = NewType("TokenIdentify", int)
TokenIdentify = str
# TokenPayload = NewType("TokenPayload", dict[str, Any])
TokenPayload = dict[str, Any]
UserIdentify = Union[int, str]

class Token(TypedDict):
    payload: TokenPayload
    uid: UserIdentify

class TokenInfo(TypedDict):
    token: RawToken
    expires_in: NotRequired[TokenExpire]

class OpaqueToken(Token):
    idf: TokenIdentify


class OpaqueTokenInfo(TokenInfo, OpaqueToken):
    pass

class JWTToken(Token):
    jti: TokenIdentify

class JWTTokenInfo(TokenInfo, JWTToken):
    pass



import asyncio

from ..mixinBase import MixinBase
from .utils import (
    validate_opaque_token,
    validate_opaque_token_info,
    validate_jwt_token,
    validate_jwt_token_info
)
from .typings import (
    TokenLimit,
    TokenKey,
    UserTokenKey,
    TokenExpire,
    MethodType,
    OpaqueToken,
    OpaqueTokenInfo,
    JWTToken,
    JWTTokenInfo,
    TokenType,
    RawToken,
    UserIdentify,
    AsyncRedis,
    AsyncPipeline,
    TokenPayload,
    Token,
    TokenInfo
)
from .utils import (
    validate_opaque_token,
    validate_opaque_token_info,
    validate_jwt_token,
    validate_jwt_token_info
)

from abc import abstractmethod
from typing import (
    Callable,
    Union,
    ParamSpec,
    TypeAlias,
    Generic,
    TypeVar,
    Iterable,
    Coroutine
)
from functools import partial
from orjson import dumps, loads

P = ParamSpec("P")
_TokenInfos = Union[OpaqueTokenInfo, JWTTokenInfo]
_Tokens = Union[OpaqueToken, JWTToken]
TI = TypeVar("TI", bound=_TokenInfos)
T = TypeVar("T", bound=_Tokens)
TokenKeyHandler = Callable[[RawToken], TokenKey]
UserTokenKeyHandler: TypeAlias = Callable[[UserIdentify], UserTokenKey]
_validator: TypeAlias = Callable[[dict], bool]
TokenValidator = _validator
TokenInfoValidator = _validator


class TokenBaseMixin(Generic[T, TI], MixinBase):
    access_token_limit: TokenLimit | None = 1
    refresh_token_limit: TokenLimit | None = 1
    access_token_expire: TokenExpire = 3600
    refresh_token_expire: TokenExpire = 129600
    user_access_token_key: str = 'user_access_token'
    user_refresh_token_key: str = 'user_refresh_token'
    access_token_key: str = 'access_token'
    refresh_token_key: str = 'refresh_token'
    oauth2_route: list[MethodType] = []
    oauth2_class: Callable = None

    def __init__(
            self,
            token: T,
            token_info: TI,
            token_validator: TokenValidator,
            token_info_validator: TokenInfoValidator
    ):
        self._Token: T = token
        self._TokenInfo: TI = token_info
        self._token_validator = token_validator
        self._token_info_validator = token_info_validator

    @abstractmethod
    def create_token(self) -> RawToken:
        pass

    @abstractmethod
    def _make_token(self, token: Token) -> TI:
        pass

    def _get_access_token_key(self, token: RawToken) -> TokenKey:
        return TokenKey(f"{self.access_token_key}/{token}")

    def _get_refresh_token_key(self, token: RawToken) -> TokenKey:
        return TokenKey(f"{self.refresh_token_key}/{token}")

    def _get_token_key_handler(self, type: TokenType) -> TokenKeyHandler:
        if type == "ACCESS":
            return self._get_access_token_key
        elif type == "REFRESH":
            return self._get_refresh_token_key

    def _get_user_access_token_key(self, idf: UserIdentify) -> UserTokenKey:
        return UserTokenKey(f"{self.user_access_token_key}/{idf}")

    def _get_user_refresh_token_key(self, idf: UserIdentify) -> UserTokenKey:
        return UserTokenKey(f"{self.user_refresh_token_key}/{idf}")

    def _get_user_token_key_handler(self, type: TokenType) -> UserTokenKeyHandler:
        if type == "ACCESS":
            return self._get_user_access_token_key
        elif type == "REFRESH":
            return self._get_user_refresh_token_key

    async def _regist_token(
            self,
            rd: AsyncRedis,
            key: UserTokenKey,
            token_key_handler: TokenKeyHandler,
            token_expire: TokenExpire,
            **kwargs
    ) -> tuple[RawToken, TokenKey]:
        """
        Redis에 Token 등록
        """
        while await rd.sadd(
                key,
                (
                        token_key := token_key_handler(
                            token := self.create_token(**kwargs)
                        )
                )
        ) == 0:
            pass
        await rd.expire(key, token_expire)
        return RawToken(token), TokenKey(token_key)

    async def _manage_user_token(
            self,
            rd: AsyncRedis,
            key: UserTokenKey
    ) -> None:
        """
        이미 지워진 토큰 들을 user token list 에서 삭제함
        """

        async def wrap(pipe: AsyncPipeline, key: UserTokenKey):
            if len((tokens := await pipe.smembers(key))) == 0:
                return
            await pipe.watch(*tokens)
            tokens_for_delete = [i[0] async for i in ((i, await pipe.exists(i)) for i in tokens) if i[1] == 0]
            if len(tokens_for_delete) == 0:
                return
            pipe.multi()
            _: AsyncPipeline = pipe.srem(key, *tokens_for_delete)

        await rd.transaction(
            partial(wrap, key=key),
            key
        )

    async def _manage_user_token_count(
            self,
            rd: AsyncRedis,
            key: UserTokenKey,
            limit: TokenLimit
    ) -> None:
        """
        토큰 갯수 제한이 있을때 user token 리스트와 토큰들의 갯수를 관리해줌
        """

        async def wrap(
                pipe: AsyncPipeline,
                key: UserTokenKey,
                limit: TokenLimit
        ) -> None:
            if len((tokens := await pipe.smembers(key))) == 0:
                return
            await pipe.watch(*tokens)
            token_count = len(tokens)

            if token_count >= limit:
                delete_count = limit - token_count + 1  # 무조건 음수
                delete_tokens = [(i, await pipe.ttl(i)) for i in tokens]
                delete_tokens.sort(key=lambda x: x[1])
                delete_tokens = delete_tokens[:delete_count]
                pipe.multi()
                _: AsyncPipeline = pipe.delete(*(d := [i[0] for i in delete_tokens]))
                _: AsyncPipeline = pipe.srem(key, *d)
            return

        await rd.transaction(
            partial(wrap, key=key, limit=limit),
            key
        )

    async def _create_type_token(
            self,
            rd: AsyncRedis,
            payload: TokenPayload,
            identify: UserIdentify,
            type: TokenType,
            **kwargs,
    ) -> TI:
        """
        Args:
            rd:
            payload:
            identify:
            type:
            **kwargs: create_token 의 인자로 넘겨줌
        """
        if type == 'ACCESS':
            token_limit = self.access_token_limit
            token_expire = self.access_token_expire
        elif type == 'REFRESH':
            token_limit = self.refresh_token_limit
            token_expire = self.refresh_token_expire
        else:
            raise AttributeError("type muste be ACCESS or REFRESH")

        get_token_key = self._get_token_key_handler(type)
        get_user_token_key = self._get_user_token_key_handler(type)

        user_token_key: UserTokenKey = get_user_token_key(identify)

        # 토큰 갯수 및 존재 여부 관리
        await self._manage_user_token(rd=rd, key=user_token_key)
        if token_limit is not None:
            await self._manage_user_token_count(rd=rd, key=user_token_key, limit=token_limit)

        # 토큰 제작
        async def transaction(
                pipe: AsyncPipeline,
                user_token_key: UserTokenKey,
                token_key: TokenKey,
                token: Token,
                token_expire: TokenExpire
        ) -> bool:
            token = dumps(token)
            if await pipe.sismember(user_token_key, token_key) == 1:
                pipe.multi()
                _: AsyncPipeline = pipe.set(token_key, token)
                _: AsyncPipeline = pipe.expire(token_key, token_expire)
                return True
            else:
                return False

        get_raw_token_and_token_key = partial(
            self._regist_token,
            rd=rd,
            key=user_token_key,
            token_key_handler=get_token_key,
            token_expire=token_expire,
            **kwargs
        )
        get_token_info = partial(self._make_token, token=Token(uid=identify, payload=payload))
        while not await rd.transaction(
                partial(
                    transaction,
                    user_token_key=user_token_key,
                    token_key=(result := await get_raw_token_and_token_key())[1],
                    token=(token_info := get_token_info()),
                    token_expire=token_expire
                ),
                *(user_token_key, result[1]),
                value_from_callable=True
        ):
            pass

        ttl = await rd.ttl(result[1])
        token_info: TI
        token_info["expires_in"] = ttl
        token_info["token"] = result[0]
        return token_info

    async def create_access_token(
            self,
            rd: AsyncRedis,
            payload: TokenPayload,
            identify: UserIdentify,
            **kwargs
    ) -> TI:
        """
        Args:
            rd:
            payload:
            identify:
            **kwargs: create token의 인자로 넘어감
        """
        return await self._create_type_token(
            rd=rd,
            payload=payload,
            type="ACCESS",
            identify=identify,
            **kwargs
        )

    async def create_refresh_token(
            self,
            rd: AsyncRedis,
            payload: TokenPayload,
            identify: UserIdentify,
            **kwargs
    ) -> TI:
        """
        Args:
            rd:
            payload:
            identify:
            **kwargs: create token의 인자로 넘어감
        """
        return await self._create_type_token(
            rd=rd,
            payload=payload,
            type="REFRESH",
            identify=identify,
            **kwargs
        )

    async def get_type_token(
            self,
            rd: AsyncRedis,
            token: RawToken,
            type: TokenType = "ACCESS",
    ) -> TI | None:
        get_token_key = self._get_token_key_handler(type)
        token_key = get_token_key(token)

        async def transaction(
                pipe: AsyncPipeline,
                raw_token: RawToken,
                key: TokenKey
        ) -> TI | None:
            if (_token := await pipe.get(key)) is None:
                await pipe.unwatch()
                return None

            token = self._Token(**loads(_token))
            if not self._token_validator(token):
                return None

            ttl = await rd.ttl(key)

            return self._TokenInfo(
                token=raw_token,
                expires_in=ttl,
                **token
            )

        return await rd.transaction(
            partial(
                transaction,
                raw_token=token,
                key=token_key,
            ),
            token_key,
            value_from_callable=True
        )

    async def get_access_token(
            self,
            rd: AsyncRedis,
            token: RawToken,
    ) -> TI | None:
        return await self.get_type_token(rd=rd, token=token, type="ACCESS")

    async def get_refresh_token(
            self,
            rd: AsyncRedis,
            token: RawToken,
    ) -> TI | None:
        return await self.get_type_token(rd=rd, token=token, type="REFRESH")

    async def _get_user_type_tokens(
            self,
            rd: AsyncRedis,
            identify: UserIdentify,
            type: TokenType = "ACCESS",
    ) -> list[TI]:
        get_token_key = self._get_token_key_handler(type)
        get_user_token_key = self._get_user_token_key_handler(type)

        key = get_user_token_key(identify)

        async def transaction(
                pipe: AsyncPipeline,
                key: UserTokenKey
        ) -> list[TI]:
            tokens = await pipe.smembers(key)
            if len(tokens) == 0:
                return []

            await pipe.watch(*tokens)

            return [self._TokenInfo(
                token=i.split('/')[1],
                expires_in=ttl,
                **token
            )
            for i in tokens if self._token_validator((token := self._Token(**loads(await pipe.get(i))))) and (ttl := await pipe.ttl(i)) > 5]

        await self._manage_user_token(rd=rd, key=key)
        res: list[TI] = await rd.transaction(
            partial(transaction, key=key),
            key,
            value_from_callable=True
        )
        res.sort(key=lambda t: t['expires_in'])

        return res

    async def get_user_access_tokens(
            self,
            rd: AsyncRedis,
            identify: UserIdentify
    ) -> list[TI]:
        return await self._get_user_type_tokens(
            rd=rd,
            identify=identify,
            type="ACCESS"
        )

    async def get_user_refresh_tokens(
            self,
            rd: AsyncRedis,
            identify: UserIdentify
    ) -> list[TI]:
        return await self._get_user_type_tokens(
            rd=rd,
            identify=identify,
            type="REFRESH"
        )

    async def _abort_user_type_token(
            self,
            rd: AsyncRedis,
            identify: UserIdentify,
            type: TokenType,
            user_tokens: list[RawToken] | None = None,
    ) -> None:
        async def transaction(
                pipe: AsyncPipeline,
                key: UserTokenKey,
                get_token_key: TokenKeyHandler,
                user_tokens: list[RawToken] | None = None
        ) -> None:
            tokens = await pipe.smembers(key)
            if len(tokens) == 0:
                return

            if user_tokens is None or len(user_tokens) == 0:
                tokens_for_delete = list(tokens)
            else:
                tokens_for_delete = [d for i in user_tokens if (d := get_token_key(i)) in tokens]

            if len(tokens_for_delete) == 0:
                return

            pipe.multi()
            _: AsyncPipeline = pipe.srem(key, *tokens_for_delete)
            _: AsyncPipeline = pipe.delete(*tokens_for_delete)

        get_token_key = self._get_token_key_handler(type)
        get_user_token_key = self._get_user_token_key_handler(type)
        user_token_key = get_user_token_key(identify)

        await rd.transaction(
            partial(transaction, key=user_token_key, get_token_key=get_token_key, user_tokens=user_tokens),
            user_token_key
        )

    async def abort_user_access_token(
            self,
            rd: AsyncRedis,
            identify: UserIdentify,
            user_tokens: RawToken | Iterable[RawToken] | None = None
    ) -> None:
        default = partial(self._abort_user_type_token, rd=rd, identify=identify, type="ACCESS")
        if user_tokens is None:
            await default()
        elif isinstance(user_tokens, Iterable):
            await default(user_tokens=[i for i in user_tokens])
        elif isinstance(user_tokens, str):
            await default(user_token=[user_tokens])
        else:
            raise ValueError('user Token error')

    async def abort_user_refresh_token(
            self,
            rd: AsyncRedis,
            identify: UserIdentify,
            user_tokens: RawToken | Iterable[RawToken] | None = None
    ) -> None:
        default = partial(self._abort_user_type_token, rd=rd, identify=identify, type="REFRESH")
        if user_tokens is None:
            await default()
        elif isinstance(user_tokens, Iterable):
            await default(user_tokens=[i for i in user_tokens])
        elif isinstance(user_tokens, str):
            await default(user_token=[user_tokens])
        else:
            raise ValueError('user Token error')

    async def abort_token(
            self,
            rd: AsyncRedis,
            token: RawToken,
    ) -> None:
        get_access_token_key = self._get_token_key_handler("ACCESS")
        get_refresh_token_key = self._get_token_key_handler("REFRESH")

        await asyncio.gather(
            rd.delete(get_access_token_key(token)),
            rd.delete(get_refresh_token_key(token))
        )

    async def _abort_type_token(
            self,
            rd: AsyncRedis,
            token: RawToken,
            type: TokenType
    ) -> None:
        token_key = self._get_token_key_handler(type)(token)
        await rd.delete(token_key)

    async def abort_access_token(
            self,
            rd: AsyncRedis,
            token: RawToken,
    ) -> None:
        await self._abort_type_token(rd=rd, token=token, type='ACCESS')

    async def abort_refresh_token(
            self,
            rd: AsyncRedis,
            token: RawToken,
    ) -> None:
        await self._abort_type_token(rd=rd, token=token, type='REFRESH')


from fastapi_namespace.mixins.token import OpaqueTokenMixin
from fastapi_namespace.mixins.token.typings import UserIdentify
from fastapi_namespace import Namespace
from fastapi_redis import FastAPIAsyncRedis
from fastapi_redis.types import AsyncRedisClient
from fastapi import Depends
from typing import Literal
import asyncio
redis = FastAPIAsyncRedis()


namespace = Namespace(
    prefix="/t",
)

@namespace.route('')
class Root(OpaqueTokenMixin):
    access_token_limit = 10
    refresh_token_limit = 10
    access_token_expire = 10
    @namespace.doc(summary="가져오기 토큰 테스트")
    async def get(
            self,
            type: Literal["ACCESS", "REFRESH"],
            identify: UserIdentify,
            rd: AsyncRedisClient = Depends(redis.get_connection),
    ):
        return await self._get_user_type_tokens(rd, identify, type)

    @namespace.doc(summary="토큰 만들기 테스트")
    async def post(
            self,
            identify: UserIdentify,
            rd: AsyncRedisClient = Depends(redis.get_connection)
    ):
        access_token, refresh_token = await asyncio.gather(
            self.create_access_token(rd=rd, identify=identify, payload={'a': "123123123"}),
            self.create_refresh_token(rd=rd, identify=identify, payload={'a': "123123123"}),
        )
        return access_token, refresh_token

    @namespace.doc(summary="토큰 삭제 테스트")
    async def delete(
            self,
            type: Literal["ACCESS", "REFRESH"],
            identify: UserIdentify,
            rd: AsyncRedisClient = Depends(redis.get_connection)
    ):
        a = await self._abort_user_type_token(rd=rd, identify=identify, type=type, user_tokens=None)











from fastapi import FastAPI
app = FastAPI()
app.include_router(namespace)


if __name__ == "__main__":
    from uvicorn import run
    run(
        app=app,
        port=23457
    )



from pydantic import TypeAdapter, ValidationError, ConfigDict
from .typings import (
    OpaqueToken as OpaqueToken_,
    OpaqueTokenInfo as OpaqueTokenInfo_,
    JWTToken as JWTToken_,
    JWTTokenInfo as JWTTokenInfo_,
)
from typing import Literal

class OpaqueToken(OpaqueToken_):
    pass
class OpaqueTokenInfo(OpaqueTokenInfo_):
    pass
class JWTToken(JWTToken_):
    pass
class JWTTokenInfo(JWTTokenInfo_):
    pass

def _default_validate(validator: TypeAdapter, data: dict) -> bool:
    try:
        validator.validate_python(data, strict=True)
        return True
    except ValidationError as e:
        return False

def validate_opaque_token(data: dict, extra: Literal['forbid', 'allow', 'ignore'] = 'forbid') -> bool:
    OpaqueToken.__pydantic_config__ = ConfigDict(extra=extra)
    OpaqueTokenValidator = TypeAdapter(OpaqueToken)
    return _default_validate(OpaqueTokenValidator, data)

def validate_opaque_token_info(data: dict, extra: Literal['forbid', 'allow', 'ignore'] = 'forbid') -> bool:
    OpaqueTokenInfo.__pydantic_config__ = ConfigDict(extra=extra)
    OpaqueTokenInfoValidator = TypeAdapter(OpaqueTokenInfo)
    return _default_validate(OpaqueTokenInfoValidator, data)

def validate_jwt_token(data: dict, extra: Literal['forbid', 'allow', 'ignore'] = 'forbid') -> bool:
    JWTToken.__pydantic_config__ = ConfigDict(extra=extra)
    JWTTokenValidator = TypeAdapter(JWTToken)
    return _default_validate(JWTTokenValidator, data)

def validate_jwt_token_info(data: dict, extra: Literal['forbid', 'allow', 'ignore'] = 'forbid') -> bool:
    JWTTokenInfo.__pydantic_config__ = ConfigDict(extra=extra)
    JWTTokenInfoValidator = TypeAdapter(JWTTokenInfo)
    return _default_validate(JWTTokenInfoValidator, data)


from setuptools import find_packages, setup

with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name='fastapi_namespace_vet1ments',
    description="For FastAPI Routing Class",
    version='0.1.10',
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires='>=3.11.0',
    author="no hong seok",
    author_email="vet1ments@naver.com",
    maintainer="no hong seok",
    maintainer_email="vet1ments@naver.com",
    project_urls={
        "Repository": "https://github.com/vet1ments/fastapi_namespace"
    },
    license="MIT",
    packages=find_packages(),
    install_requires=[
        "fastapi>=0.108.0"
    ],
    tests_require=[
        "uvicorn>=0.9.0",
        "fastapi-redis-vet1ments>=0.2.6"
    ],
)


from fastapi_namespace.resource import Resource
from fastapi_namespace.utils import delete_none
from fastapi_namespace.typings import MethodType
from typing import Iterable
from abc import ABCMeta
from fastapi.params import Depends

class _Meta(type):
    def __new__(mcs, name, bases, namespace):
        global_dependencies: Iterable[Depends] = getattr(bases[0], 'global_dependencies', None)
        if global_dependencies is not None:
            _global_dependencies = namespace.get('global_dependencies', [])
            global_dependencies = [*global_dependencies, *_global_dependencies]

        get_dependencies: Iterable[Depends] = getattr(bases[0], 'get_dependencies', None)
        if get_dependencies is not None:
            _get_dependencies = namespace.get('get_dependencies', [])
            get_dependencies = [*get_dependencies, *_get_dependencies]

        post_dependencies: Iterable[Depends] = getattr(bases[0], 'post_dependencies', None)
        if post_dependencies is not None:
            _post_dependencies = namespace.get('post_dependencies', [])
            post_dependencies = [*post_dependencies, *_post_dependencies]

        put_dependencies: Iterable[Depends] = getattr(bases[0], 'put_dependencies', None)
        if put_dependencies is not None:
            _put_dependencies = namespace.get('put_dependencies', [])
            put_dependencies = [*put_dependencies, *_put_dependencies]

        delete_dependencies: Iterable[Depends] = getattr(bases[0], 'delete_dependencies', None)
        if delete_dependencies is not None:
            _delete_dependencies = namespace.get('delete_dependencies', [])
            delete_dependencies = [*delete_dependencies, *_delete_dependencies]

        options_dependencies: Iterable[Depends] = getattr(bases[0], 'options_dependencies', None)
        if options_dependencies is not None:
            _options_dependencies = namespace.get('options_dependencies', [])
            options_dependencies = [*options_dependencies, *_options_dependencies]

        head_dependencies: Iterable[Depends] = getattr(bases[0], 'head_dependencies', None)
        if head_dependencies is not None:
            _head_dependencies = namespace.get('head_dependencies', [])
            head_dependencies = [*head_dependencies, *_head_dependencies]

        patch_dependencies: Iterable[Depends] = getattr(bases[0], 'patch_dependencies', None)
        if patch_dependencies is not None:
            _patch_dependencies = namespace.get('patch_dependencies', [])
            patch_dependencies = [*patch_dependencies, *_patch_dependencies]

        trace_dependencies: Iterable[Depends] = getattr(bases[0], 'trace_dependencies', None)
        if trace_dependencies is not None:
            _trace_dependencies = namespace.get('trace_dependencies', [])
            trace_dependencies = [*trace_dependencies, *_trace_dependencies]

        namespace.update(delete_none({
            'global_dependencies': global_dependencies,
            'get_dependencies': get_dependencies,
            'post_dependencies': post_dependencies,
            'put_dependencies': put_dependencies,
            'delete_dependencies': delete_dependencies,
            'options_dependencies': options_dependencies,
            'head_dependencies': head_dependencies,
            'patch_dependencies': patch_dependencies,
            'trace_dependencies': trace_dependencies,
        }))
        return type.__new__(mcs, name, bases, namespace)

class __Meta(_Meta, ABCMeta):
    pass

class MixinBase(Resource, metaclass=__Meta):
    global_dependencies: Iterable[Depends]
    get_dependencies: Iterable[Depends]
    post_dependencies: Iterable[Depends]
    put_dependencies: Iterable[Depends]
    delete_dependencies: Iterable[Depends]
    options_dependencies: Iterable[Depends]
    head_dependencies: Iterable[Depends]
    patch_dependencies: Iterable[Depends]
    trace_dependencies: Iterable[Depends]

    def _add_depends(self, type_: MethodType, depends: Depends) -> None:
        assert isinstance(depends, Depends), "Depends must be a Depends"
        key = f'{type_}_dependencies'
        dependencies = getattr(self, key, [])
        dependencies.append(depends)
        setattr(self, key, dependencies)

    def add_global_depends(self, depends: Depends) -> None:
        return self._add_depends('global', depends)

    def add_get_depends(self, depends: Depends) -> None:
        return self._add_depends('get', depends)

    def add_post_depends(self, depends: Depends) -> None:
        return self._add_depends('post', depends)

    def add_put_depends(self, depends: Depends) -> None:
        return self._add_depends('put', depends)

    def add_delete_depends(self, depends: Depends) -> None:
        return self._add_depends('delete', depends)

    def add_options_depends(self, depends: Depends) -> None:
        return self._add_depends('options', depends)

    def add_head_depends(self, depends: Depends) -> None:
        return self._add_depends('head', depends)

    def add_patch_depends(self, depends: Depends) -> None:
        return self._add_depends('patch', depends)

    def add_trace_depends(self, depends: Depends) -> None:
        return self._add_depends('trace', depends)